/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Id: UnixHTTPURLInputStream.cpp 568078 2007-08-21 11:43:25Z amassari $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#if !defined(XML_BEOS)
  #include <netinet/in.h>
  #include <arpa/inet.h>
#endif
#include <netdb.h>
#include <errno.h>

#include <xercesc/util/XMLNetAccessor.hpp>
#include <xercesc/util/NetAccessors/Socket/UnixHTTPURLInputStream.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLExceptMsgs.hpp>
#include <xercesc/util/Janitor.hpp>
#include <xercesc/util/XMLUniDefs.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/util/TranscodingException.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/Base64.hpp>

XERCES_CPP_NAMESPACE_BEGIN

class SocketJanitor
{
public:
    // -----------------------------------------------------------------------
    //  Constructors and Destructor
    // -----------------------------------------------------------------------
    SocketJanitor(int* toDelete) : fData(toDelete) {}
    ~SocketJanitor() { reset(); }

    int* get() const { return fData; }
    int* release() { int* p = fData; fData = 0; return p; }

    void reset(int* p = 0)
    {
        if(fData) {
            shutdown(*fData, 2);
            close(*fData);
        }
        fData = p;
    }
    bool isDataNull() { return (fData == 0); }

private :
    // -----------------------------------------------------------------------
    //  Unimplemented constructors and operators
    // -----------------------------------------------------------------------
    SocketJanitor();
    SocketJanitor(const SocketJanitor&);
    SocketJanitor& operator=(const SocketJanitor&);

    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fData
    //      This is the pointer to the socket that must be closed when 
    //      this object is destroyed.
    // -----------------------------------------------------------------------
    int*  fData;
};

class TranscodeStr
{
public:
    TranscodeStr(const XMLCh *in, XMLTranscoder* trans,
                 MemoryManager *manager = XMLPlatformUtils::fgMemoryManager)
        : fString(0),
          fBytesWritten(0),
          fMemoryManager(manager)
    {
        if(in) {
            unsigned int len = XMLString::stringLen(in) + 1;

            unsigned int allocSize = len * sizeof(XMLCh);
            fString = (unsigned char*)fMemoryManager->allocate(allocSize);

            unsigned int charsRead;

            unsigned int charsDone = 0;

            while(true) {
                fBytesWritten += trans->transcodeTo(in + charsDone, len - charsDone,
                                                    fString + fBytesWritten, allocSize - fBytesWritten,
                                                    charsRead, XMLTranscoder::UnRep_Throw);
                charsDone += charsRead;

                if(charsDone == len) break;

                allocSize *= 2;
                unsigned char *newBuf = (unsigned char*)fMemoryManager->allocate(allocSize);
                memcpy(newBuf, fString, fBytesWritten);
                fMemoryManager->deallocate(fString);
                fString = newBuf;
            }
        }
    }

    ~TranscodeStr()
    {
        if(fString)
            fMemoryManager->deallocate(fString);
    }

    const unsigned char *str() const
    {
        return fString;
    }

    unsigned int len() const
    {
        return fBytesWritten;
    }

private:
    unsigned char *fString;
    unsigned int fBytesWritten;
    MemoryManager *fMemoryManager;
};

static const char *CRLF = "\r\n";

void UnixHTTPURLInputStream::createHTTPRequest(const XMLURL &urlSource, const XMLNetHTTPInfo *httpInfo, CharBuffer &buffer)
{
    static const char *GET = "GET ";
    static const char *PUT = "PUT ";
    static const char *POST = "POST ";
    static const char *HTTP10 = " HTTP/1.0\r\n";
    static const char *HOST = "Host: ";
    static const char *AUTHORIZATION = "Authorization: Basic ";
    static const char *COLON = ":";

    XMLTransService::Codes failReason;
    const unsigned int blockSize = 2048;

    XMLTranscoder* trans = XMLPlatformUtils::fgTransService->makeNewTranscoderFor("ISO8859-1", failReason, blockSize, fMemoryManager);
    Janitor<XMLTranscoder> janTrans(trans);

    TranscodeStr hostName(urlSource.getHost(), trans, fMemoryManager);
    TranscodeStr path(urlSource.getPath(), trans, fMemoryManager);
    TranscodeStr fragment(urlSource.getFragment(), trans, fMemoryManager);
    TranscodeStr query(urlSource.getQuery(), trans, fMemoryManager);

    // Build up the http GET command to send to the server.
    // To do:  We should really support http 1.1.  This implementation
    //         is weak.
    if(httpInfo) {
        switch(httpInfo->fHTTPMethod) {
        case XMLNetHTTPInfo::GET:   buffer.append(GET); break;
        case XMLNetHTTPInfo::PUT:   buffer.append(PUT); break;
        case XMLNetHTTPInfo::POST:  buffer.append(POST); break;
        }
    }
    else {
        buffer.append(GET);
    }

    if(path.str() != 0) {
        buffer.append((char*)path.str());
    }
    else {
        buffer.append("/");
    }

    if(query.str() != 0) {
        buffer.append("?");
        buffer.append((char*)query.str());
    }

    if(fragment.str() != 0) {
        buffer.append((char*)fragment.str());
    }
    buffer.append(HTTP10);

    buffer.append(HOST);
    buffer.append((char*)hostName.str());
    if(urlSource.getPortNum() != 80)
    {
        buffer.append(COLON);
        buffer.appendDecimalNumber(urlSource.getPortNum());
    }
    buffer.append(CRLF);

    const XMLCh *username = urlSource.getUser();
    const XMLCh *password = urlSource.getPassword();
    if(username && password) {
        XMLBuffer userPassBuf(256, fMemoryManager);
        userPassBuf.append(username);
        userPassBuf.append(chColon);
        userPassBuf.append(password);

        TranscodeStr userPass(userPassBuf.getRawBuffer(), trans, fMemoryManager);

        unsigned int len;
        XMLByte* encodedData = Base64::encode((XMLByte*)userPass.str(), userPass.len() - 1, &len, fMemoryManager);
        ArrayJanitor<XMLByte> janBuf2(encodedData, fMemoryManager);

        if(encodedData) {
            // HTTP doesn't want the 0x0A separating the data in chunks of 76 chars per line
            XMLByte* authData = (XMLByte*)fMemoryManager->allocate((len+1)*sizeof(XMLByte));
            ArrayJanitor<XMLByte> janBuf(authData, fMemoryManager);
            XMLByte *cursor = authData;
            for(unsigned int i = 0; i < len; ++i)
                if(encodedData[i] != chLF)
                    *cursor++ = encodedData[i];
            *cursor++ = 0;
            buffer.append(AUTHORIZATION);
            buffer.append((char*)authData);
            buffer.append(CRLF);
        }
    }

    if(httpInfo && httpInfo->fHeaders)
        buffer.append(httpInfo->fHeaders, httpInfo->fHeadersLen);

    buffer.append(CRLF);
}

int UnixHTTPURLInputStream::parseResponseStatus() const
{
    char *p = const_cast<char*>(strstr(fBuffer.getRawBuffer(), "HTTP"));
    if(p == 0) return -1;

    p = strchr(p, chSpace);
    if(p == 0) return -1;
  
    return atoi(p);
}

XMLCh *UnixHTTPURLInputStream::findHeader(const char *name) const
{
    int len = strlen(name);

    char *p = const_cast<char*>(strstr(fBuffer.getRawBuffer(), name));
    while(p != 0) {
        if(*(p - 1) == '\n' &&
            *(p + len) == ':' &&
            *(p + len + 1) == ' ') {

            p += len + 2;
            
            char *endP = strstr(p, CRLF);
            if(endP == 0) {
                for(endP = p; *endP != 0; ++endP);
            }

            char tmp = *endP;
            *endP = 0;

            XMLCh *value = XMLString::transcode(p, fMemoryManager);
            *endP = tmp;
            return value;
        }

        p = strstr(p + 1, name);
    }

    return 0;
}

bool UnixHTTPURLInputStream::send(const char *buf, unsigned int len)
{
    unsigned int done = 0;
    int ret;

    while(done < len) {
        ret = ::send(fSocket, buf + done, len - done, 0);
        if(ret == -1) return false;
        done += ret;
    }

    return true;
}

UnixHTTPURLInputStream::UnixHTTPURLInputStream(const XMLURL& urlSource, const XMLNetHTTPInfo* httpInfo/*=0*/)
      : fSocket(0)
      , fBytesProcessed(0)
      , fBuffer(1023, urlSource.getMemoryManager())
      , fContentType(0)
      , fMemoryManager(urlSource.getMemoryManager())
{
    //
    //  Constants in ASCII to send/check in the HTTP request/response
    //

    static const char *CRLF2X = "\r\n\r\n";
    static const char *LF2X = "\n\n";

    //
    //  Convert the hostName to the platform's code page for gethostbyname and
    //  inet_addr functions.
    //

    const XMLCh*        hostName = urlSource.getHost();
    char*               hostNameAsCharStar = XMLString::transcode(hostName, fMemoryManager);
    ArrayJanitor<char>  janHostNameAsCharStar(hostNameAsCharStar, fMemoryManager);

    XMLURL url(urlSource);
    int redirectCount = 0;
    SocketJanitor janSock(0);
    
    do {
        //
        // Set up a socket.
        //

        struct hostent *hostEntPtr = 0;
        struct sockaddr_in sa;

        // Use the hostName in the local code page ....
        if((hostEntPtr = gethostbyname(hostNameAsCharStar)) == NULL)
        {
            unsigned long  numAddress = inet_addr(hostNameAsCharStar);
            if ((hostEntPtr =
                    gethostbyaddr((char *) &numAddress,
                        sizeof(unsigned long), AF_INET)) == NULL)
            {
                ThrowXMLwithMemMgr1(NetAccessorException,
                    XMLExcepts::NetAcc_TargetResolution, hostName, fMemoryManager);
            }
        }

        memset(&sa, '\0', sizeof(sockaddr_in));  // iSeries fix ??
        memcpy((void *) &sa.sin_addr,
            (const void *) hostEntPtr->h_addr, hostEntPtr->h_length);
        sa.sin_family = hostEntPtr->h_addrtype;
        sa.sin_port = htons((unsigned short)url.getPortNum());

        janSock.reset();
        fSocket = socket(hostEntPtr->h_addrtype, SOCK_STREAM, 0);
        if(fSocket < 0)
        {
            ThrowXMLwithMemMgr1(NetAccessorException,
                XMLExcepts::NetAcc_CreateSocket, url.getURLText(), fMemoryManager);
        }
        janSock.reset(&fSocket);

        if(connect(fSocket, (struct sockaddr *) &sa, sizeof(sa)) < 0)
        {
            ThrowXMLwithMemMgr1(NetAccessorException,
                XMLExcepts::NetAcc_ConnSocket, url.getURLText(), fMemoryManager);
        }


        // The port is open and ready to go.
        // Build up the http GET command to send to the server.
        CharBuffer requestBuffer(1023, fMemoryManager);
        createHTTPRequest(url, httpInfo, requestBuffer);

        // Send the http request
        if(!send(requestBuffer.getRawBuffer(), requestBuffer.getLen())) {
            ThrowXMLwithMemMgr1(NetAccessorException,
                XMLExcepts::NetAcc_WriteSocket, url.getURLText(), fMemoryManager);
        }

        if(httpInfo && httpInfo->fPayload) {
            if(!send(httpInfo->fPayload, httpInfo->fPayloadLen)) {
                ThrowXMLwithMemMgr1(NetAccessorException,
                    XMLExcepts::NetAcc_WriteSocket, url.getURLText(), fMemoryManager);
            }
        }

        //
        // get the response, check the http header for errors from the server.
        //
        char tmpBuf[1024];
        int ret;

        fBuffer.reset();
        while(true) {
            ret = ::recv(fSocket, tmpBuf, sizeof(tmpBuf), 0);
            if(ret == -1) {
                ThrowXMLwithMemMgr1(NetAccessorException, XMLExcepts::NetAcc_ReadSocket, url.getURLText(), fMemoryManager);
            }

            fBuffer.append(tmpBuf, ret);

            fBufferPos = strstr(fBuffer.getRawBuffer(), CRLF2X);
            if(fBufferPos != 0) {
                fBufferPos += 4;
                *(fBufferPos - 2) = 0;
                break;
            }

            fBufferPos = strstr(fBuffer.getRawBuffer(), LF2X);
            if(fBufferPos != 0) {
                fBufferPos += 2;
                *(fBufferPos - 1) = 0;
                break;
            }
        }

        int status = parseResponseStatus();
        if(status < 0) {
            ThrowXMLwithMemMgr1(NetAccessorException, XMLExcepts::NetAcc_ReadSocket, url.getURLText(), fMemoryManager);
        }

        if(status == 200) {
            // HTTP 200 OK response means we're done.
            // We're done
            break;
        }
        // a 3xx response means there was an HTTP redirect
        else if(status >= 300 && status <= 307) {
            redirectCount++;

            XMLCh *newURLString = findHeader("Location");
            ArrayJanitor<XMLCh> janNewURLString(newURLString, fMemoryManager);

            XMLURL newURL(fMemoryManager);
            newURL.setURL(url, newURLString);
            if(newURL.getProtocol() != XMLURL::HTTP) {
                ThrowXMLwithMemMgr1(NetAccessorException, XMLExcepts::File_CouldNotOpenFile, newURL.getURLText(), fMemoryManager);
            }

            url = newURL;

            janHostNameAsCharStar.release();
            hostNameAsCharStar = XMLString::transcode(newURL.getHost(), fMemoryManager);
            janHostNameAsCharStar.reset(hostNameAsCharStar, fMemoryManager);
        }
        else {    		
            // Most likely a 404 Not Found error.
            ThrowXMLwithMemMgr1(NetAccessorException, XMLExcepts::File_CouldNotOpenFile, url.getURLText(), fMemoryManager);
        }
    } while(redirectCount < 6);

    janSock.release();
}


UnixHTTPURLInputStream::~UnixHTTPURLInputStream()
{
    shutdown(fSocket, 2);
    close(fSocket);

    if(fContentType) fMemoryManager->deallocate(fContentType);
}

const XMLCh *UnixHTTPURLInputStream::getContentType() const
{
    if(fContentType == 0) {
        // mutable
        const_cast<XMLCh*&>(fContentType) = findHeader("Content-Type");
    }
    return fContentType;
}

unsigned int UnixHTTPURLInputStream::readBytes(XMLByte* const    toFill
                                      , const unsigned int    maxToRead)
{
    unsigned int len = fBuffer.getRawBuffer() + fBuffer.getLen() - fBufferPos;
    if(len > 0)
    {
        // If there's any data left over in the buffer into which we first
        //   read from the server (to get the http header), return that.
        if (len > maxToRead)
            len = maxToRead;
        memcpy(toFill, fBufferPos, len);
        fBufferPos += len;
    }
    else
    {
        // There was no data in the local buffer.
        // Read some from the socket, straight into our caller's buffer.
        //
        len = read(fSocket, (void *) toFill, maxToRead);
        if (len == (unsigned int)-1)
        {
            ThrowXMLwithMemMgr(NetAccessorException, XMLExcepts::NetAcc_ReadSocket, fMemoryManager);
        }
    }

    fBytesProcessed += len;
    return len;
}

XERCES_CPP_NAMESPACE_END

